"""This pkg aims to implement serveral filtering methods for (un)directed
graphs.

Edge filtering methods allows to extract the backbone of a graph or sampling the
most important edges. You can use edge filtering methods as a preprocessing step
aiming to improve the performance/results of graph algorithms or to turn a graph
visualtzation more asthetic.

**See the example below for a simple usage of the package.**
```python
import networkx as nx
import edgeseraser as ee

g = nx.erdos_renyi_graph(100, 0.1)
ee.noise_score.filter_nx_graph(g, field=None)

g # filtered graph
```

## Available methods and details

| Method | Description | suitable for | limitations/restrictions/details |
| --- | --- |--- | --- |
| [Noise Score] | Filters edges with high noise score. Paper:[1]|Directed, Undirected, Weighted | Very good and fast! [4] |
| [Disparity] | Dirichlet process filter (stick-breaking) Paper:[2] |  Directed, Undirected, Weighted |There are some criticism regarding the use in undirected graphs[3]|

[1]: https://arxiv.org/abs/1701.07336
[2]: https://arxiv.org/abs/0904.
[3]: https://arxiv.org/abs/2101.00863
[4]: https://www.michelecoscia.com/?p=1236
[Noise Score]: /api_docs/#edgeseraser.noise_score
[Disparity]: /api_docs/#edgeseraser.disparity

"""

__author__ = """Bruno Messias"""
__email__ = "devmessias@gmail.com"
__version__ = "0.1.0"



import sys
import warnings

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

import numpy as np
import scipy.sparse as sp  # type: ignore

warnings.simplefilter("ignore", FutureWarning)


def get_disparity_integral(norm_weight, degree):
    """
    calculate the significance (alpha) for the disparity filter
    """
    disparity_integral = ((1.0 - norm_weight) ** degree) / (
        (degree - 1.0) * (norm_weight - 1.0)
    )

    return disparity_integral


def stick_break_scores(
    w_degree: np.ndarray, degree: np.ndarray, edges: np.ndarray, weights: np.ndarray
) -> np.ndarray:
    """
    Calculate the stick-breaking scores for each edge.

    Args:
        w_degree: np.array
            edge weighted degree
        degree: np.array
            degree of each vertex
        edges: np.array
            edges of the graph
        weights: np.array
            edge weights

    Returns:
        np.array:
            **alphas** stick-breaking scores for each edge
    """

    st = w_degree[edges[:, 0]]
    degree = degree[edges[:, 0]]
    w_degree = w_degree[edges[:, 0]]

    ids_w0 = np.argwhere(st == 0)
    ids_d1 = np.argwhere(degree < 2)
    st[ids_w0] = 1
    degree[ids_d1] = 10
    st = weights / st
    st[st == 1.0] -= 10e-4
    alphas = 1.0 - (
        (degree - 1.0)
        * (get_disparity_integral(st, degree) - get_disparity_integral(0.0, degree))
    )

    alphas[ids_d1] = 0.0
    alphas[ids_w0] = 0.0
    return alphas


def cond_stick_edges2erase(alphas: np.ndarray, thresh: float = 0.1) -> np.ndarray:
    """
    Args:
        alphas: np.array
            edge scores
        thresh: float
            Between 0 and 1.
    Returns:
        np.array:
            indices of edges to be erased

    """
    ids2erase = np.argwhere(alphas > thresh).flatten()
    return ids2erase


def filter_generic_graph(
    num_vertices: int,
    edges: np.ndarray,
    weights: np.ndarray,
    cond: Literal["or", "both", "out", "in"] = "or",
) -> np.ndarray:
    """
    Args:
        num_vertices: int
            number ofvertices
        edges: np.array
            edges
        weights: np.array
            edge weights
        cond: str
            "out", "in", "both", "or"
    Returns:
        np.array:
            **alphas** edge scores

    """
    w_adj = sp.csr_matrix((weights, edges.T), shape=(num_vertices, num_vertices))
    adj = sp.csr_matrix(
        (np.ones(edges.shape[0]), edges.T), shape=(num_vertices, num_vertices)
    )

    calc_degree = lambda x, i: np.asarray(x.sum(axis=i)).flatten().astype(np.float64)
    w_degree_out = calc_degree(w_adj, 0)
    degree_out = calc_degree(adj, 0)
    w_degree_in = calc_degree(w_adj, 1)
    degree_in = calc_degree(adj, 1)
    if cond == "out":
        alphas = stick_break_scores(w_degree_out, degree_out, edges, weights)
    elif cond == "in":
        alphas = stick_break_scores(w_degree_in, degree_in, edges, weights)
    else:
        alphas_out = stick_break_scores(w_degree_out, degree_out, edges, weights)
        alphas_in = stick_break_scores(w_degree_in, degree_in, edges, weights)
        if cond == "both":
            alphas = np.maximum(alphas_out, alphas_in)
        elif cond == "or":
            alphas = np.minimum(alphas_out, alphas_in)

    return alphas


def filter_nx_graph(
    g,
    thresh: float = 0.5,
    cond: Literal["or", "both", "out", "in"] = "or",
    field: str = "weight",
) -> None:
    """Filter edges from a networkx graph using the disparity filter.
    (Dirichet proccess)

    Parameters:
        g: networkx.Graph
            graph to be filtered
        thresh: float
            Between 0 and 1.
        cond: str
            "out", "in", "both", "or"

    """
    assert thresh > 0.0 and thresh < 1.0, "thresh must be between 0 and 1"

    num_vertices = g.number_of_nodes()
    if field is None:
        edges = np.array([[u, v, 1.0] for u, v in g.edges()])
    else:
        edges = np.array([[u, v, d[field]] for u, v, d in g.edges(data=True)])

    weights = edges[:, 2].astype(np.float64)
    edges = edges[:, :2].astype(np.int64)

    alphas = filter_generic_graph(num_vertices, edges, weights, cond=cond)
    ids2erase = cond_stick_edges2erase(alphas, thresh=thresh)
    g.remove_edges_from([(e[0], e[1]) for e in edges[ids2erase]])


def filter_ig_graph(
    g,
    thresh: float = 0.5,
    cond: Literal["or", "both", "out", "in"] = "or",
    field: str = "weight",
) -> None:
    """Filter edges from a igraph instance using the disparity filter.
    (Dirichet proccess)

    Parameters:
        g: igraph.Graph
            graph to be filtered
        thresh: float
            Between 0 and 1.
        cond: str
            "out", "in", "both", "or"

    """
    assert thresh > 0.0 and thresh < 1.0, "thresh must be between 0 and 1"

    num_vertices = g.vcount()
    edges = np.array(g.get_edgelist())
    if field is None:
        weights = np.ones(edges.shape[0])
    else:
        weights = np.array(g.es[field]).astype(np.float64)

    alphas = filter_generic_graph(num_vertices, edges, weights, cond=cond)
    ids2erase = cond_stick_edges2erase(alphas, thresh=thresh)
    g.delete_edges(ids2erase)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['edgeseraser']

package_data = \
{'': ['*']}

install_requires = \
['igraph', 'networkx', 'numpy', 'scipy', 'typing_extensions']

setup_kwargs = {
    'name': 'edgeseraser',
    'version': '0.3.0',
    'description': 'A short description of the project',
    'long_description': '# Edges Eraser\n\n[![PyPI](https://img.shields.io/pypi/v/edgeseraser?style=flat-square)](https://pypi.python.org/pypi/edgeseraser/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/edgeseraser?style=flat-square)](https://pypi.python.org/pypi/edgeseraser/)\n[![PyPI - License](https://img.shields.io/pypi/l/edgeseraser?style=flat-square)](https://pypi.python.org/pypi/edgeseraser/)\n\n[![Coookiecutter - Wolt](https://img.shields.io/badge/cookiecutter-Wolt-00c2e8?style=flat-square&logo=cookiecutter&logoColor=D4AA00&link=https://github.com/woltapp/wolt-python-package-cookiecutter)](https://github.com/woltapp/wolt-python-package-cookiecutter)\n\n\n---\n\n**Documentation**: [https://devmessias.github.io/edgeseraser](https://devmessias.github.io/edgeseraser)\n\n**Source Code**: [https://github.com/devmessias/edgeseraser](https://github.com/devmessias/edgeseraser)\n\n**PyPI**: [https://pypi.org/project/edgeseraser/](https://pypi.org/project/edgeseraser/)\n\n---\n\n## What is Edges Eraser?\nThis pkg aims to implement serveral filtering methods for (un)directed graphs.\n\nEdge filtering methods allows to extract the backbone of a graph or sampling the most important edges. You can use edge filtering methods as a preprocessing step aiming to improve the performance/results of graph algorithms or to turn a graph visualtzation more asthetic.\n\n\n## Example\n```python\nimport networkx as nx\nimport edgeseraser as ee\n\ng = nx.erdos_renyi_graph(100, 0.1)\nee.noise_score.filter_nx_graph(g, field=None)\n\ng # filtered graph\n```\n\n## Available methods and details\n\n| Method | Description | suitable for | limitations/restrictions/details |\n| --- | --- |--- | --- |\n| [Noise Score] | Filters edges with high noise score. Paper:[1]|Directed, Undirected, Weighted | Very good and fast! [4] |\n| [Disparity] | Dirichlet process filter (stick-breaking) Paper:[2] |  Directed, Undirected, Weighted |There are some criticism regarding the use in undirected graphs[3]|\n\n[1]: https://arxiv.org/abs/1701.07336\n[2]: https://arxiv.org/abs/0904.\n[3]: https://arxiv.org/abs/2101.00863\n[4]: https://www.michelecoscia.com/?p=1236\n[Noise Score]: /api_docs/#edgeseraser.noise_score\n[Disparity]: /api_docs/#edgeseraser.disparity\n\n\n## Installation\n\n```sh\npip install edgeseraser\n```\n\n## Development\n\n* Clone/Fork this repository\n\n```sh\ngit clone https://github.com/devmessias/edgeseraser\n```\n\n* Requirements:\n  * [Poetry](https://python-poetry.org/)\n  * Python 3.7+\n\n```sh\nmake install\nmake init\n```\n\n### Testing\n\n```sh\nmake test\n```\n\nTo run the static analysis, use the following command:\n```sh\nmake mypy\n```\n\n### Documentation\n\nThe documentation is automatically generated from the content of the [docs directory](./docs) and from the docstrings\n of the public signatures of the source code. The documentation is updated and published as a [Github project page\n ](https://pages.github.com/) automatically as part each release.\n\nTo see the current state of the documentation in your browser, use the following command:\n```sh\nmake docs-serve\n```\nThe above command will start a local server on port 8000. Any changes to\nthe documentation and docstrings will be automatically reflected in your browser.\n\n### Pre-commit\n\nPre-commit hooks run all the auto-formatters (e.g. `black`, `isort`), linters (e.g. `mypy`, `flake8`), and other quality\n checks to make sure the changeset is in good shape before a commit/push happens.\n\n\nIf you want e.g. want to run all checks manually for all files:\n\n```sh\nmake pre-commit\n```\n',
    'author': 'Bruno Messias',
    'author_email': 'devmessias@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://devmessias.github.io/edgeseraser',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)



import warnings
from typing import Tuple, TypeVar

import networkx as nx  # type: ignore
import numpy as np
import scipy.sparse as sp  # type: ignore

warnings.simplefilter("ignore", FutureWarning)

NpOrFloat = TypeVar("NpOrFloat", np.ndarray, float)


def get_noise_score(
    st_u: NpOrFloat, st_v: NpOrFloat, vol: float, w: NpOrFloat
) -> Tuple[NpOrFloat, NpOrFloat]:
    """
    Get noise score for each edge

    Args:
        st_u: np.array
            strength for each vertex
        st_v: np.array
            strength for each vertex
        vol: float
            volume of the graph
        w: np.array
            edge weights

    Returns:
        (np.array, np.array):
        -   **scores_uv**
            Noise score for each edge.
        -   **std_uv**
            standard deviation of noise score for each edge

    """
    st_prod_uv = np.multiply(st_u, st_v)

    mean_prior_prob = st_prod_uv / (vol) ** 2.0
    kappa = vol / st_prod_uv
    score = (kappa * w - 1) / (kappa * w + 1)
    var_prior_prob = (
        (1 / vol**2.0)
        * (st_prod_uv * (vol - st_u) * (vol - st_v))
        / (vol**2 * (vol - 1))
    )

    alpha_prior = (mean_prior_prob**2 / var_prior_prob) * (
        1 - mean_prior_prob
    ) - mean_prior_prob

    beta_prior = (
        (mean_prior_prob / var_prior_prob) * (1 - mean_prior_prob**2.0)
        - 1
        + mean_prior_prob
    )

    alpha_post = alpha_prior + w
    beta_post = vol - w + beta_prior
    expected_puv = alpha_post / (alpha_post + beta_post)
    variance_uv = expected_puv * (1 - expected_puv) * vol
    d = 1 / (st_prod_uv) - vol * ((st_u + st_v) / st_prod_uv**2)
    variance_cuv = variance_uv * (2 * (kappa + w * d) / (kappa * w + 1) ** 2.0) ** 2.0
    sdev_cuv = variance_cuv**0.5

    return score, sdev_cuv


def filter_generic_graph(
    w_degree: np.ndarray, edges: np.ndarray, weights: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute noise corrected edge weights for a sparse graph.

    Args:
        w_degree: np.ndarray
            Weight degree of each vertex.
        edges: np.array
            Edges of the graph.
        weights: np.array
            Edge weights of the graph.

    Returns:
        (np.array, np.array):
        -   **scores_uv**
            Noise corrected edge weights.
        -   **std_uv**
            Standard deviation of noise corrected edge weights.

    """

    vol = w_degree.sum()

    st_u = w_degree[edges[:, 0]]
    st_v = w_degree[edges[:, 1]]
    scores = st_u * st_v
    ids_0 = np.argwhere(scores == 0)
    st_u[ids_0] = 1
    st_v[ids_0] = 1
    scores_uv, std_uv = get_noise_score(st_u, st_v, vol, weights)
    scores_uv[ids_0] = 0
    std_uv[ids_0] = 0.0
    return scores_uv, std_uv


def cond_noise_edges2erase(
    scores_uv: np.ndarray, std_uv: np.ndarray, thresh: float = 1.28
) -> np.ndarray:
    """Filter edges with high noise score.

    Args:
        scores_uv: np.array
            edge scores
        std_uv: np.array
            edge standard deviations
        thresh: float
            >Since this is roughly equivalent to a one-tailed test of
            statistical significance, common values of δ are 1.28, 1.64, and
            2.32, which approximate p-values of 0.1, 0.05, and 0.0

    Returns:
        np.array:
        indices of edges to be erased
    """
    ids2erase = np.argwhere(scores_uv <= thresh * std_uv).flatten()
    return ids2erase


def filter_nx_graph(g, thresh: float = 1.28, field: str = "weight") -> None:
    """Filter edge with high noise score from a networkx graph.

    Args:
        g: networkx.Graph
            Graph to be filtered.
        thresh: float
            >Since this is roughly equivalent to a one-tailed test of
            statistical significance, common values of δ are 1.28, 1.64, and
            2.32, which approximate p-values of 0.1, 0.05, and 0.0
        field: str
            Edge field to be used for filtering.

    Example:
        ```python
        import networkx as nx
        import edgeseraser as ee

        g = nx.erdos_renyi_graph(100, 0.1)
        ee.noise_score.filter_nx_graph(g, field=None)

        g # filtered graph
        ```

    """

    w_adj = nx.adjacency_matrix(g)
    w_degree = np.asarray(w_adj.sum(axis=1)).flatten().astype(np.float64)
    if field is None:
        edges = np.array([[u, v, 1.0] for u, v in g.edges()])
    else:
        edges = np.array([[u, v, d[field]] for u, v, d in g.edges(data=True)])
    weights = edges[:, 2].astype(np.float64)
    edges = edges[:, :2].astype(np.int64)

    scores_uv, std_uv = filter_generic_graph(w_degree, edges, weights)

    ids2erase = cond_noise_edges2erase(scores_uv, std_uv, thresh=thresh)
    g.remove_edges_from([(e[0], e[1]) for e in edges[ids2erase]])


def filter_ig_graph(g, thresh: float = 1.28, field: str = "weight") -> None:
    """Filter edge with high noise score from a igraph graph.

    Args:
        g: igraph.Graph
            Graph to be filtered.
        thresh: float
            >Since this is roughly equivalent to a one-tailed test of
            statistical significance, common values of δ are 1.28, 1.64, and
            2.32, which approximate p-values of 0.1, 0.05, and 0.0
        field: str
            Edge field to be used for filtering.

    Example:
        ```python
        import networkx as nx
        import edgeseraser as ee

        g = nx.erdos_renyi_graph(100, 0.1)
        ee.noise_score.filter_nx_graph(g, field=None)

        g # filtered graph
        ```

    """
    num_vertices = g.vcount()
    edges = np.array(g.get_edgelist())
    if field is None:
        weights = np.ones(edges.shape[0])
    else:
        weights = np.array(g.es[field]).astype(np.float64)

    w_adj = sp.csr_matrix((weights, edges.T), shape=(num_vertices, num_vertices))
    w_degree = np.asarray(w_adj.sum(axis=1)).flatten().astype(np.float64)

    scores_uv, std_uv = filter_generic_graph(w_degree, edges, weights)

    ids2erase = cond_noise_edges2erase(scores_uv, std_uv, thresh=thresh)
    g.delete_edges(ids2erase)
